//-----------------------------------------------------------------------------
/*
Programmed by Olivier Sangala
http://olivier.sangala.free.fr
olivier_sangala@hotmail.com
*/
//-----------------------------------------------------------------------------

// C Source File
// Created 14/06/2005; 19:08:54

#include <tigcclib.h>

#include <extgraph.h>


//---------------------------------------------------------------------------

#define NUM_OBJECTS_MAX (8)
#define NUM_COLUMNS_MAX (2*NUM_OBJECTS_MAX)

//#define NUM_OBJECTS (8)
#define NUM_COLUMNS (2*NUM_OBJECTS)

//---------------------------------------------------------------------------

enum {TL_EMPTY,TL_CURSOR,TL_WHEEL,TL_LINE};

enum {REC_OK,REC_ABORTED};

typedef struct
{
	char Columns_Used[NUM_COLUMNS_MAX];
	char Objects_Pos[NUM_OBJECTS_MAX];
}
TBoard;



//---------------------------------------------------------------------------

void Reset(void);

void Solve_Main(void);
int Solve_Rec(int ObjectID);

int IsSolution(void);

void DrawScreen(void);

//---------------------------------------------------------------------------

TBoard*Board;

int NUM_OBJECTS=8;
int ObjectCurPos=0;

int solve_mode=0;
int solve_cnt_solutions=0;


//---------------------------------------------------------------------------

unsigned short data_tiles[40] =
{
    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 
    0x0000, 0x0000, 0xFFC0, 0xFFC0, 0xFFC0, 0xFFC0, 0xFFC0, 0xFFC0, 
    0xFFC0, 0xFFC0, 0xFFC0, 0xFFC0, 0x0C00, 0x3300, 0x4C80, 0x5280, 
    0xA140, 0xA140, 0x5280, 0x4C80, 0x3300, 0x0C00, 0x0000, 0x0000, 
    0x0000, 0x0000, 0xFFC0, 0xFFC0, 0x0000, 0x0000, 0x0000, 0x0000, 
} ;

//---------------------------------------------------------------------------


void Reset(void)
{
	ObjectCurPos=0;
	memset(Board,0,sizeof(TBoard));
}
//---------------------------------------------------------------------------

void Solve_Main(void)
{
	int key;
	
	clrscr();
	printf("F1 : Get Num Solutions \n");
	printf("F2 : List Solutions \n");
	printf("F3 : Exit \n");
	
	while ((key=ngetchx())!=KEY_F1 && key!=KEY_F2 && key!=KEY_F3);
	
	if (key==KEY_F3)
	return;
	
	solve_mode=(key==KEY_F2?1:0);
	solve_cnt_solutions=0;
	
	printf("Processing ... \n");
	
	Reset();
	int r=Solve_Rec(0);
	
	// CLEAR KEYSTROKES
	GKeyFlush();
	
	
	if (r==REC_ABORTED)
	{
		printf("Process aborted \n");
		ngetchx();
	}
	
	if (solve_mode==0)
	{
		printf("Num Solutions = %d \n",solve_cnt_solutions);
		ngetchx();
	}
	
	printf("Press any key ... \n");
	ngetchx();
	
}
//---------------------------------------------------------------------------


int Solve_Rec(int ObjectID)
{
	
	int p;
	
	for (p=0;p<=(NUM_COLUMNS-ObjectID-2);p++)
	if (Board->Columns_Used[p]==0 && Board->Columns_Used[p+ObjectID+1]==0)
	{
		Board->Objects_Pos[ObjectID]=p;
		
		Board->Columns_Used[p]=1;
		Board->Columns_Used[p+ObjectID+1]=1;
		
		if (ObjectID==(NUM_OBJECTS-1))
		{
			
			
			if (solve_mode==1)
			{
				printf("#%d : ",solve_cnt_solutions);
				int a;
				for (a=0;a<NUM_OBJECTS;a++)
				printf("%d ",Board->Objects_Pos[a]);
				printf(" \n");
				
				// ABORT CASE
				
				int key=ngetchx();
				
				if (key==KEY_ESC)
				return REC_ABORTED;
				
			}
			
			solve_cnt_solutions++;
		}
		else
		{
			int r = Solve_Rec(ObjectID+1);
			
			if (r==REC_ABORTED)
			return r;
		}
		
		Board->Columns_Used[p]=0;
		Board->Columns_Used[p+ObjectID+1]=0;
	}
	
	
	return REC_OK;
	
	
}


//---------------------------------------------------------------------------

int IsSolution(void)
{
	char UsedColumns_List[NUM_COLUMNS_MAX];
	memset(UsedColumns_List,0,NUM_COLUMNS_MAX);
	
	// MAKE USED COLUMNS LIST
	
	int ObjectID;
	
	for (ObjectID=0;ObjectID<NUM_OBJECTS;ObjectID++)
	{
		int WheelPos1 = Board->Objects_Pos[ObjectID];
		int WheelPos2 = Board->Objects_Pos[ObjectID]+ObjectID+1;
		
		UsedColumns_List[WheelPos1]|=1;
		UsedColumns_List[WheelPos2]|=1;
		
	}
	
	// CHECK IF ANY COLUMN IS NOT USED
	
	int p;
	for (p=0;p<NUM_COLUMNS;p++)
	if (!UsedColumns_List[p])
	return 0;
	
	// ELSE, IS SOLUTION
	return 1;
}
//---------------------------------------------------------------------------


void DrawScreen(void)
{
	
	int p,ObjectID;
	
	ClrScr();
	
	// TILES
	
	for (ObjectID=0;ObjectID<NUM_OBJECTS;ObjectID++)
	for (p=0;p<NUM_COLUMNS;p++)
	{
		int WheelPos1 = Board->Objects_Pos[ObjectID];
		int WheelPos2 = Board->Objects_Pos[ObjectID]+ObjectID+1;
		
		int TileID=((p==WheelPos1 || p==WheelPos2)?TL_WHEEL:
		(p>WheelPos1 && p<WheelPos2)?TL_LINE:TL_EMPTY);
		
		unsigned short*src=data_tiles+10*TileID;
		
		if (TileID!=TL_EMPTY)
		Sprite16_OR(10*p,10*ObjectID,10,src,LCD_MEM);
		
	}
	
	// MESSAGES
	
	char buffer[16];
	sprintf(buffer,"#%d",NUM_OBJECTS);
	
	DrawStr(0,92,buffer,A_NORMAL);
	
	if (IsSolution())
	DrawStr(30,92,"You Win!",A_NORMAL);
	
}
//---------------------------------------------------------------------------


// Main Function
void _main(void)
{
	// Place your code here.
	
	TBoard BoardA;
	Board=&BoardA;
	
	
	clrscr();
	FontSetSys(0);
	printf("Wheemy 1.0 \n");
	printf("Programmed by Olivier Sangala \n");
	printf("http://olivier.sangala.free.fr \n");
	printf("olivier_sangala@hotmail.com \n");
	FontSetSys(1);
	ngetchx();
	
	
	Reset();
	DrawScreen();
	
	for (;;)
	{
		
		unsigned short*src=data_tiles+10*TL_CURSOR;
		int a;
		
		for (a=0;a<NUM_COLUMNS;a++)
		Sprite16_XOR(10*a,10*ObjectCurPos,10,src,LCD_MEM);
		
		int key=ngetchx();
		
		for (a=0;a<NUM_COLUMNS;a++)
		Sprite16_XOR(10*a,10*ObjectCurPos,10,src,LCD_MEM);
		
		
		
		if (key==KEY_UP)
		ObjectCurPos=(ObjectCurPos==0?(NUM_OBJECTS-1):ObjectCurPos-1);
		else if (key==KEY_DOWN)
		ObjectCurPos=(ObjectCurPos==(NUM_OBJECTS-1)?0:ObjectCurPos+1);
		
		
		else if (key==KEY_LEFT)
		{
			if (Board->Objects_Pos[ObjectCurPos]>0)
			{
				Board->Objects_Pos[ObjectCurPos]--;
				DrawScreen();
			}
		}
		
		else if (key==KEY_RIGHT)
		{
			if (Board->Objects_Pos[ObjectCurPos]<(NUM_COLUMNS-ObjectCurPos-2))
			{
				Board->Objects_Pos[ObjectCurPos]++;
				DrawScreen();
			}
		}
		
		
		else if (key==KEY_F1)
		{
			Solve_Main();
			
			DrawScreen();
		}
		
		else if (key==KEY_F2)
		{
			NUM_OBJECTS=(NUM_OBJECTS==8?3:NUM_OBJECTS+1);
			
			Reset();
			DrawScreen();
		}
		
		else if (key==KEY_ESC)
		break;
		
	}
	
	
}



